// SELECTOR function


/** selector function:
 *    Args:
 *      pattern: a string that contains comma-separated #id or .classname or tagname or [attributename=value]
 *      scope: an HTMLElement
 *    Returns:
 *      array: list of matching HTMLElements
 *      HTMLElement: if exactly one id as pattern
 */
function $(pattern, scope){
  if(!scope){var scope = document};
  var selectors = pattern.split(/\s*?,\s*?/);
  var results = [];
  for(var s=0; s<selectors.length; s++){ selector = selectors[s];
    if(selector.charAt(0)=="#"){
      var e = scope.getElementById(selector.slice(1));
      if(selectors.length==1){return e} else{results.push(e)};
    } else if(selector.charAt(0)=="."){
      if(document.getElementsByClassName){ results = results.concat( Array.prototype.slice.call(scope.getElementsByClassName(selector.slice(1))) )
      } else{ var cs = scope.getElementsByTagName("*"); if(cs.length==0){continue};
      var regexp = new RegExp('\\b'+selector.slice(1)+'\\b', 'gi');
      for(var i=0;i<cs.length;i++){ if(cs[i].className.search(regexp)!=-1){results.push(cs[i])} };
      }
    } else if(selector.charAt(0)=="["&&selector.slice(-1)=="]"){
      var attribute = selector.match(/[^\[\]\=]+/)[0];
      var value = selector.match(/\=[^\]]+(?=\])/); if(value!=null) value = value[0].slice(1);
      var cs = scope.getElementsByTagName("*"); if(cs.length==0){continue};
      for(var i=0;i<cs.length;i++){
        var val = cs[i].getAttribute(attribute);
        if(val!=null&&((value!=null)?(val==value):true)){results.push(cs[i])}
      };
    } else{
      var cs = scope.getElementsByTagName(selector);
      for(var i=0;i<cs.length;i++){results.push(cs[i])};
    }
  }
  return results;
}




// PROTOTYPE functions


if(typeof(Array.prototype.each) === "undefined"){
  Array.prototype.each = function(callback){
    for (var i =  0; i < this.length; i++){
      callback(this[i]);
    }
  }
}

if(typeof(Array.prototype.each_with_index) === "undefined"){
  Array.prototype.each_with_index = function(callback){
    for (var i =  0; i < this.length; i++){
      callback(this[i], i);
    }
  }
}


if(typeof(Array.prototype.collect) === "undefined"){
  Array.prototype.collect = function(callback){
    var results = [];
    for (var i =  0; i < this.length; i++){
      results[i] = callback(this[i]);
    }
    return results
  }
}


if(typeof(Array.prototype.inject) === "undefined"){
  Array.prototype.inject = function(startvalue, func){
    for (var i =  0; i < this.length; i++){
      startvalue = func(startvalue, this[i]);
    }
    return startvalue;
  }
}


if(typeof(String.prototype.trim) === "undefined"){
  String.prototype.trim = function(){
    return String(this).replace(/^\s+|\s+$/g, '');
  };
}


if(typeof(String.prototype.ljust) === "undefined"){
  String.prototype.ljust = function( width, padding ){
    padding = padding || " ";
    padding = padding.substr( 0, 1 );
    if( this.length < width ){
      return this + padding.repeat( width - this.length );
    } else{
      return this;
    }
  }
}


if(typeof(String.prototype.rjust) === "undefined"){
  String.prototype.rjust = function( width, padding ){
    padding = padding || " ";
    padding = padding.substr( 0, 1 );
    if( this.length < width ){
      return padding.repeat( width - this.length ) + this;
    } else{
      return this;
    }
  }
}


if(typeof(String.prototype.repeat) === "undefined"){
  String.prototype.repeat = function( num ){
    for( var i = 0, buf = ""; i < num; i++ ){buf += this};
    return buf;
  }
}


if(typeof(String.prototype.lines) === "undefined"){
  String.prototype.lines = function(){
    var lines = this.match(/\r?\n/g);
    return !lines? 1 : lines.length + 1;
  }
}


if(typeof(String.prototype.toCamelCase) === "undefined"){
  String.prototype.toCamelCase = function(){
    return this.replace(/^-+|-+$/g,'').replace( /(-|^)([a-z])/g, function(t,a,b) { return (a=="")? b.toLowerCase() : b.toUpperCase(); } );
  }
}




/* STYLE Functions */


// Show element.
function show(elem){
  elem.style.display = elem.orig_display || "block";
}


// Hide element.
function hide(elem){
  // remember original display property because it could be block,inline-block,inline etc.
  elem.orig_display = elem.display;
  elem.style.display = "none";
}


function hasClass(elem, className){ return new RegExp("(^|\\b)"+className+"(\\b|$)").test(elem.className) }


function addClass(elem, className){if(!hasClass(elem,className)) elem.className += (elem.className ? ' ' : '')+className}


function removeClass(elem, className){ var r = new RegExp("(^\\s*|\\s*\\b)"+className+"(\\b|$)"); elem.className = elem.className.replace(r,"")}


function toggleClass(elem, className, value){ if(value==null){var value = !hasClass(elem, className)}; if(value){ addClass(elem, className) } else{ removeClass(elem, className) }; }


function changeCssRule(selector, properties){
  var sheets = document.styleSheets;
  for(var i=0; i<sheets.length; i++){
    var rules = sheets[i].cssRules || sheets[i].rules; // Mozille || IE
    for(var j=0; j<rules.length; j++){
      if(rules[j].selectorText == selector){
        for(prop in properties){ rules[j].style[prop] = properties[prop] };
      }
    }
  }
  // if a rule with matching selector doesn't exist, create a new one
  if(sheets[0].insertRule){
    var props = [];
    for(prop in properties){ props.push(prop+": "+properties[prop]) };
    sheets[0].insertRule(selector+" {" + props.join("; ") + "}", 0);
  } else if(sheets[0].addRule){ // IE
    for(prop in properties){
      sheets[0].addRule(selector, prop + ":" + properties[prop], 0);
    }
  };
}


function supportsCss(arg){
  if(typeof arg == 'string'){ var selector = arg;
    // test string as selector
    var stylesheet = document.body.appendChild( document.createElement('style') );
    var sheet = document.styleSheets[document.styleSheets.length-1];
    var rules = sheet.cssRules || sheet.rules;
    var l = rules.length;
    try{
      if(sheet.insertRule){ sheet.insertRule(selector+" {}",0) }
      else{ sheet.addRule(selector, "a-property",0) };
      return rules.length>l;
    } catch(e){
      return rules.length>l;
    } finally{
      document.body.removeChild(stylesheet);
    }
  } else{
    // test hash as property/value pairs
    var elem = document.createElement('elem');
    var result = true;
    for(property in arg){
      // test if property exists
      var prop = property.toCamelCase();
      if(result) result = ( prop in elem.style );
      // optionally test if value is valid
      if(result && arg[property]){
        try{ elem.style[prop] = arg[property] }
        catch(e){ result = false }
      };
    }
    return result;
  }
}




// EVENT functions


function bindEvent(elem, eventname, callback){
  if(elem.addEventListener){  
    elem.addEventListener(eventname, callback, false);
    return true; 
  } else if (elem.attachEvent){ // IE
    return elem.attachEvent("on"+eventname, callback);  
  }
}


function triggerEvent(elem, eventname){
  if(document.createEvent){
    var e = document.createEvent('HTMLEvents');
    e.initEvent(eventname, true, false);
    elem.dispatchEvent(e);
  } else if(document.createEventObject){ // IE
    if(elem == window || elem == document){ elem = document.body }
    elem.fireEvent("on"+eventname);
  }
}




// SELECTION functions


// http://stackoverflow.com/a/263796
function getCaretPosition(elem){ 
  if(elem.selectionStart){ 
    return elem.selectionStart; 
  } else if(document.selection){ 
    elem.focus(); 
    var r = document.selection.createRange(); 
    if(r == null){ 
      return 0; 
    }
    var re = elem.createTextRange(), 
        rc = re.duplicate(); 
    re.moveToBookmark(r.getBookmark()); 
    rc.setEndPoint('EndToStart', re); 
    return rc.text.length; 
  }  
  return 0; 
}


// http://www.blog.vishalon.net/index.php/javascript-getting-and-setting-caret-position-in-textarea/
function setCaretPosition(elem, pos){
	if(elem.setSelectionRange){
		elem.focus();
		elem.setSelectionRange(pos,pos);
	}
	else if (elem.createTextRange){
		var range = elem.createTextRange();
		range.collapse(true);
		range.moveEnd('character', pos);
		range.moveStart('character', pos);
		range.select();
	}
}


function getSelectionText(){
  var text = false;
  if(window.getSelection) { text = window.getSelection().toString(); }
  else if(document.getSelection) { text = document.getSelection().toString(); }
  else{
    var selection = document.selection && document.selection.createRange();
    if(selection.text){ text = selection.text; }
  }
  return text;
}


function getSelectionHTML(){
  var userSelection;
  if(window.getSelection){
    // W3C Ranges
    userSelection = window.getSelection();
    // Get the range:
    if(userSelection.rangeCount == 0){ return false }
    if(userSelection.getRangeAt)
      var range = userSelection.getRangeAt(0);
    else{
      var range = document.createRange();
      range.setStart(userSelection.anchorNode, userSelection.anchorOffset);
      range.setEnd(userSelection.focusNode, userSelection.focusOffset);
    }
    // And the HTML:
    var clonedSelection = range.cloneContents();
    var div = document.createElement('div');
    div.appendChild(clonedSelection);
    return div.innerHTML;
  } else if(document.selection){
    // Explorer selection, return the HTML
    userSelection = document.selection.createRange();
    return userSelection.htmlText;
  } else{
    return false;
  }
};


// http://stackoverflow.com/questions/401593/textarea-selection
function elemGetSelection(elem){
  //Mozilla and DOM 3.0
  if("selectionStart" in elem){
    var l = elem.selectionEnd - elem.selectionStart;
    return { start: elem.selectionStart, end: elem.selectionEnd, length: l, text: elem.value.substr(elem.selectionStart, l) };
  }
  //IE
  else if(document.selection){
    elem.focus();
    var r = document.selection.createRange();
    var tr = elem.createTextRange();
    var tr2 = tr.duplicate();
    tr2.moveToBookmark(r.getBookmark());
    tr.setEndPoint('EndToStart',tr2);
    if (r == null || tr == null) return { start: elem.value.length, end: elem.value.length, length: 0, text: '' };
    var text_part = r.text.replace(/[\r\n]/g,'.'); //for some reason IE doesn't always count the \n and \r in the length
    var text_whole = elem.value.replace(/[\r\n]/g,'.');
    var the_start = text_whole.indexOf(text_part,tr.text.length);
    return { start: the_start, end: the_start + text_part.length, length: text_part.length, text: r.text };
  }
  //Browser not supported
  else return { start: elem.value.length, end: elem.value.length, length: 0, text: '' };
}


function elemReplaceSelection(elem, replace_str){
  selection = elemGetSelection(elem);
  var start_pos = selection.start;
  var end_pos = start_pos + replace_str.length;
  elem.value = elem.value.substr(0, start_pos) + replace_str + elem.value.substr(selection.end, elem.value.length);
  //elemSetSelection(elem, start_pos, end_pos);
  return {start: start_pos, end: end_pos, length: replace_str.length, text: replace_str};
}


function elemSetSelection(elem, start_pos,end_pos){
  //Mozilla and DOM 3.0
  if("selectionStart" in elem){
    elem.focus();
    elem.selectionStart = start_pos;
    elem.selectionEnd = end_pos;
  }
  //IE
  else if(document.selection){
    elem.focus();
    var tr = elem.createTextRange();
    //Fix IE from counting the newline characters as two seperate characters
    var stop_it = start_pos;
    for (i=0; i < stop_it; i++) if( elem.value[i].search(/[\r\n]/) != -1 ) start_pos = start_pos - 0.5;
    stop_it = end_pos;
    for (i=0; i < stop_it; i++) if( elem.value[i].search(/[\r\n]/) != -1 ) end_pos = end_pos - 0.5;
    tr.moveEnd('textedit',-1);
    tr.moveStart('character',start_pos);
    tr.moveEnd('character',end_pos - start_pos);
    tr.select();
  }
  return elemGetSelection(elem);
}




// USER INTERFACE functions


//  Resize the height of the webdialog so that everything fits well.
//  Measures the HTML document's width and height and send them to Ruby.
function resizeHeight() {
  window.location = 'skp:resizeHeight@['+($("body")[0].clientWidth|$("body")[0].scrollWidth)+','+($("body")[0].clientHeight|$("body")[0].scrollHeight)+']';
}


/** Toggle button widget.
 *    Args:
 *      elem: an HTML element
 *      fx: an optional function to execute when the value changes
 *      def: a default value (true/false or the value of a radio element)
 */
function ToggleButton(elem, fx, def){
  elem.pressed = false;
  var set = function(v){
    elem.pressed = !v;
    toggleClass(elem, "pressed", !v)
    fx(v);
  }
  elem.onclick = function(){ set(elem.pressed) }
  if(def!=null){set(def)}
}


/** Make an element resizable with the window.
 *    Args:
 *      elem: an element
 *      percent: how much the element's size should follow the window size, 
 *        optional, useful when multiple Resizables (Number y or Array[x,y])
 *      align: where to keep the scroll position aligned, 
 *        optional, 0 = top/left, 1 = bottom/right
 */
function Resizable(elem, percent, align){
  if(!percent){ var percent = [100,100] } else if(typeof percent == "Number"){ percent = [null,percent] }
  if(!align){ var align = [null,null] } else if(typeof align == "Boolean"){ align = [null,align] }
  var min = 1;
  var body = $("body")[0];
  body.previous_width = (body.clientWidth||body.scrollWidth);
  var _this = this;
  _this.resize = function(){
    var scrollH = (align[0]==0)? elem.scrollLeft : elem.scrollWidth - elem.offsetWidth - elem.scrollLeft;
    var scrollV = (align[1]==0)? elem.scrollTop : elem.scrollHeight - elem.offsetHeight - elem.scrollTop;
    if(percent[0] != null && percent[0] != 0){
      var dw = (window.innerWidth||document.documentElement.clientWidth) - body.previous_width;
      var w = elem.offsetWidth + percent[0]/100 * dw;
      elem.style.width = ( w<min?min:w ) + "px";
    };
    if(percent[1] != null && percent[1] != 0){
      var dh = (window.innerHeight||document.documentElement.clientHeight) - (body.clientHeight||body.scrollHeight);
      var h = elem.offsetHeight + percent[1]/100 * dh;
      elem.style.height = ( h<min?min:h ) + "px";
    };
    if(align[0] != null){
      elem.scrollLeft = (align[0]==0)? scrollH : elem.scrollWidth - elem.offsetWidth - scrollH;
    };
    if(align[1] != null){
      elem.scrollTop = (align[1]==0)? scrollV : elem.scrollHeight - elem.offsetHeight - scrollV;
    };
    body.previous_width = (body.clientWidth||body.scrollWidth);
  }
  bindEvent(window, "resize", _this.resize);
}


/** Collapsible Panel widget.
 *    Args:
 *      elem: an HTML element containing a .collapsible-panel-legend and a .collapsible-panel-content
 *        Clicking on the legend hides/shows the content
 *      def: a default value (true/false)
 */
function CollapsePanel(elem, def){
  var _this = this;
  legend = $(".collapsible-panel-legend", elem)[0];
  var useCollapseArrow = !supportsCss(":before"); // whether to add the arrow symbol with javascript
  legend.innerHTML = "\<label\>" + (useCollapseArrow? "\<span class='arrow-down'\>▼\<\/span\>\<span class='arrow-right'\>►\<\/span\>" : "") + legend.innerHTML + "\<\/label\>";
  var label = $("label",legend)[0];
  elem.collapsed = false;
  var set = function(v){
    elem.collapsed = !v;
    toggleClass(elem, "collapsed", !v);
    resizeHeight();
    _this.ontoggle(v);
  }
  _this.ontoggle = function(){};
  elem.toggle = function(){ set(elem.collapsed) };
  bindEvent(label, "mouseup", elem.toggle);
  // set default
  if(def!=null){set(def)}
}


/** Navigatable History for a input[type=text] or textarea
 *    Navigate up or down using the arrow keys ↑↓, 
 *    press enter to submit the inserted text,
 *    press shift-enter for a line-break.
 *    Args:
 *      elem: an input[type=text] or textarea
 */
function History(elem){
  var _this = elem;
  if(!_this.history){ _this.history = [""] };
  _this.selected = 0;
  elem.updateHeight = function(){
    var lines = elem.value.lines();
    if(lines > config.inputMaxLines){ lines = config.inputMaxLines; elem.style.overflow = "auto" } else if(lines == 0){ lines = 1 } else{ elem.style.overflow = "hidden" };
    elem.style.height = 1.25 * lines + "em"; //  IMPORTANT: insert lineHeight in em if lineHeight!=1em
    triggerEvent(window, "resize");
  }
  var caret_pos = 0;
  // keyboard button events to navigate through the history
  bindEvent(elem, "keydown", function(event){
    if(!event){ var event = window.event };

    // enter
    if(event.keyCode == 13 && !event.shiftKey){
      if(typeof(elem.submit) !== "undefined"){
        // prevent line break
        if (event.preventDefault) { // Firefox
          event.stopPropagation();
          event.preventDefault();
        } else { // IE
          event.cancelBubble = true;
          event.returnValue = false;
        }
        // only if the entry is different from the previous one, we add it to the history
        if(elem.value != _this.history[1]){
          _this.history[0] = elem.value;
          if(_this.history.length > config.historyMaxLength ){ _this.history.pop() }
          _this.history.unshift(""); // a new, empty current entry
        }
        // submit it and reset
        elem.submit(elem.value);
        elem.value = "";
        _this.selected = 0;
      }
    }
    // shift-enter
    else if(event.keyCode == 13 && event.shiftKey){}
    // arrow up
    else if(event.keyCode == 38){
      var caret_in_first_line = elem.value.slice(0,getCaretPosition(elem)).lines() == 1;
      // do something if we are not at the last entry
      if(_this.selected < _this.history.length-1 && caret_in_first_line){
        if(_this.selected == 0){ _this.history[0] = elem.value };
        _this.selected++;
        elem.value = _this.history[_this.selected];
        // set the caret in the first line at the remembered position or at the end
        var new_caret_pos = Math.min( caret_pos, elem.value.search(/\r?\n|$/) );
        window.setTimeout(function(){setCaretPosition(elem, new_caret_pos)}, 0);
      }
    }
    // arrow down
    else if(event.keyCode == 40){
      var lines = elem.value.lines();
      var caret_in_last_line = elem.value.slice(0,getCaretPosition(elem)).lines() == lines;
      // do something only if we are not at the first (current) entry
      if(_this.selected > 0 && caret_in_last_line){
        _this.selected--;
        elem.value = _this.history[_this.selected];
        // set the caret in the last line at the remembered position
        var new_caret_pos = elem.value.search(/[^\r?\n]*$/) + caret_pos;
        window.setTimeout(function(){setCaretPosition(elem, new_caret_pos)}, 0);
      };
    }
    // any other key
    else{
      // remember the position of the caret in its line
      window.setTimeout(function(){
        caret_pos = getCaretPosition(elem) - elem.value.slice(0,getCaretPosition(elem)).search(/[^\r?\n]*$/);
      }, 0);
    }
    // update height of the input field
    window.setTimeout(function(){elem.updateHeight()}, 1);
  });
  bindEvent(elem, "mousedown", function(event){
    // remember the position of the caret in the current line
    window.setTimeout(function(){
      caret_pos = getCaretPosition(elem) - elem.value.slice(0,getCaretPosition(elem)).search(/[^\r?\n]*$/);
    }, 0);
  });
}


/** Manages items in the output field
 *    Args:
 *      elem: a block element (div)
 *    Methods:
 *      add: adds a new entry (string, type)
 *      shift: removes the oldest entry
 *      clear: removes all entries
 *      filter: shows/hides categories or entries (elem, state)
 */
function Output(elem){
  var table = document.createElement("table");
  elem.appendChild(table);
  var tbody = document.createElement("tbody");
  table.appendChild(tbody);
  _this = elem;
  _this.record = [];
  _this.createNewRow = true;

  // add a new item
  _this.add = function(string, type, time){
    if(!type){ var type = "other" };
    // keep the number of items below the maximum limit
    if(_this.record.length > config.outputMaxLength){
      _this.shift();
    }
    // make sure the text is a string and prevent code insertion (replace <>)
    var text = String(string).replace(/\</g, "&lt;").replace(/\>/g, "&gt;");
    // create a new item and a new table row
    if(_this.createNewRow){
      var d = (time)? new Date(time) : new Date();
      var time = String(d.getHours()).rjust(2,"0")+":"+String(d.getMinutes()).rjust(2,"0")+":"+String(d.getSeconds()).rjust(2,"0")+"."+String(d.getMilliseconds()).rjust(3,"0");
      var item = {"time": time, "type": type, "text": text};
      _this.record.push(item);
      _this.newRow(item)
    // or attach to the previous item and row
    } else{
      _this.record[_this.record.length-1].text += text;
      _this.attach(text);
    }
    // puts/print messages are always attached to the previous entry (createNewRow = false)
    _this.createNewRow = (type == "ruby")? false : true
  };

  // remove oldest item
  _this.shift = function(){
    _this.record.shift();
    var scrollV = elem.scrollHeight - elem.offsetHeight - elem.scrollTop;
    tbody.removeChild(tbody.firstChild);
    // preserve the scroll position from bottom
    elem.scrollTop = elem.scrollHeight - elem.offsetHeight - scrollV;
  };

  // clear the output
  _this.clear = function(){
    _this.record = [];
    for(var i = tbody.rows.length-1; i >= 0; i--){ tbody.deleteRow(i) }
  }

  // redraw the complete output again
  _this.redraw = function(){
    for(var i = tbody.rows.length-1; i >= 0; i--){ tbody.deleteRow(i) }
    _this.record.each(function(item){_this.newRow(item)});
  }

  // create a new row
  _this.newRow = function(item){
    var scrollV = elem.scrollHeight - elem.offsetHeight - elem.scrollTop;
    var lines = item.text.lines();
    var line_numbers = ""; if(lines > 1){ for(var i=1;i<=lines;i++){ line_numbers += i+"\r\n"} };
    var promptString = ""; 
    switch(item.type){
      case("input"): promptString = "&gt;"; break;
      // case("result"): promptString = "=&gt;";
      case("error"): promptString = "!"; break;
    };
    var text = item.text.replace(/\r\n/g,"\n"); //.replace(/(\r?\n)/g, "<br>"); // replace(/(\r?\n)/g, "$1<br>");
    // create the html of the row (workaround for Internet Explorer table.innerHTML bug)
    var tempNode = document.createElement('div');
    tempNode.innerHTML = "<table><tbody>" +
      "<tr class='" + item.type + "'>" +
      "<td class='time' unselectable='on'>" + item.time + "</td>" +
      "<td class='prompt' unselectable='on'>" + promptString + "</td>" +
      "<td class='line-numbers'><pre unselectable='on'>" + (lines>1? line_numbers : "") + "</pre></td>" +
      "<td class='text userSelect' unselectable='off'><pre class='ruby'><code class='ruby'>" + text + "</code></pre></td>" +
      "</tr></tbody></table>";
    tbody.appendChild(tempNode.firstChild.firstChild.firstChild); // div.table.tbody.tr
    // remember the code
    var lastCodeTag = $("code", tbody.lastChild)[0];
    lastCodeTag.parentNode.parentNode.code = text;
    // highlight the code
    if(hljs && config.showHighLight && lastCodeTag && item.type!="error"){ hljs.highlightBlock( lastCodeTag, null, false ) };
    // preserve the scroll position from bottom
    elem.scrollTop = elem.scrollHeight - elem.offsetHeight - scrollV;
  };

  // attach to the last row
  _this.attach = function(text){
    var scrollV = elem.scrollHeight - elem.offsetHeight - elem.scrollTop;
    var lastCodeTag = $("code", tbody.lastChild)[0];
    // change the code
    lastCodeTag.parentNode.parentNode.code = lastCodeTag.parentNode.parentNode.code + text.replace(/\r\n/g,"\n");
    // change the html
    lastCodeTag.innerHTML = lastCodeTag.parentNode.parentNode.code;
    // highlight the code
    if(hljs && config.showHighLight && lastCodeTag){ hljs.highlightBlock( lastCodeTag, null, false ) };
    // preserve the scroll position from bottom
    elem.scrollTop = elem.scrollHeight - elem.offsetHeight - scrollV;
  };

  // filter messages or columns by tag
  _this.filter = function(type, state){
    var scrollV = elem.scrollHeight - elem.offsetHeight - elem.scrollTop;
    if(state==true){
      changeCssRule("."+type, {"display": ""});
    } else{
      changeCssRule("."+type, {"display": "none"});
    }
    // preserve the scroll position from bottom
    elem.scrollTop = elem.scrollHeight - elem.offsetHeight - scrollV;
  }
}



