=begin

Copyright 2012, Andreas Eisenbarth
All Rights Reserved

Permission to use, copy, modify, and distribute this software for 
any purpose and without fee is hereby granted, provided that the above
copyright notice appear in all copies.

THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

Name:         ae_Console.rb
Author:       Andreas Eisenbarth
Description:  
Usage:        menu Window → Ruby Console (WebDialog)
Version:      0.6.3
Date:         19.06.2012

=end


module AE; end


class AE::Console < $stdout.class # Sketchup::Console # TODO

  #{#  CLASS VARIABLES
  #
    @@dir = File.dirname(__FILE__) # this plugin's folder
    
    @@open = false
    @@instance = nil
    
    @@console_dlg = nil
    
    @@command  = nil
    @@topmenu  = nil
    @@menuitem = nil
    
    @@debug = false
  #
  #}#


  #{#  CONSTANTS
  #
    OSX = ( Object::RUBY_PLATFORM =~ /(darwin)/i ? true : false ) unless defined?(Module.nesting[0]::OSX)
    WINE = ( ((!ENV["WINEPREFIX"].nil? || !ENV["WINELOADER"].nil? || File.exists?("Z:/usr/bin/wine")) && !OSX)? true : false ) unless defined?(Module.nesting[0]::WINE)
    WIN = ( !OSX && !WINE ) unless defined?(Module.nesting[0]::WIN)
  #
  #}#


  #{#  CLASS METHODS
  #
    #{ close()
    #
    #  Cleans up, releases instance, calls GC
    #
    def self.close
      return unless @@console_dlg
      #
      UI.start_timer(0.3,false) {
        @@console_dlg.close() if @@console_dlg.visible?
        Kernel.sleep(0.5)
        @@instance.instance_variable_set(:@console_dlg,nil)
        @@console_dlg = nil
        @@instance = nil
        $stdout = $old_stdout
        $stderr = $old_stderr
        GC.start
      }
      #
    end #}

    #{ command()
    #
    #  Provide a handle to the command, in case
    #  some other script wants to add it to a toolbar.
    #
    def self.command
      @@command
    end #}

    #{ debug accessors
    #
    def self.debug
      @@debug
    end

    def self.debug?
      @@debug
    end

    def self.debug=(arg)
      @@debug =( arg ? true : false )
    end
    #}

    def self.open(instance)
      @@instance = instance
      if @@console_dlg.nil? # nil if not yet instanced.
        @@console_dlg = instance.method(:create).call # also opens it.
      else
        # The WebDialog instance exists, just open it.
        instance.open()
      end
    end
    
    def self.open?
      return @@open # @console_dlg && @console_dlg.visible?
    end
  #
  #}#


  #{#  INSTANCE METHODS
  #
    private
    
    def initialize
      @console_dlg = nil
      # the context for local variables
      context = Proc.new{}
      @binding = context.binding
      #
      # cwh = Sketchup.read_default("Plugins_ae", "Console_Size")
      # if cwh.nil? # first run
        # @width,@height = [345,400]
        # Sketchup.write_default("Plugins_ae", "Console_Size",[345,400])
      # else
        # @width,@height = cwh
      # end
      #
      super
      #
    end

    protected

    def create()
    
      # create a webdialog
      @console_dlg = UI::WebDialog.new("Ruby Console (WebDialog)",
             false, "AE_Console", 345,400,300,300, true)
             
      ## let Sketchup track size & position
      
      @console_dlg.set_file( File.join(@@dir, "Console.htm") )
      @console_dlg.set_background_color( @console_dlg.get_default_dialog_color() )
      
      # callbacks
      @console_dlg.add_action_callback("init"){|@console_dlg, param|
        defaults = load_defaults
        # restore escape characters
        defaults = defaults.gsub(/\&/,'\\').gsub(/\\\\amp;/, '&amp;').gsub(/\\amp;/, '&')
        @console_dlg.execute_script("initialize(#{defaults})")
      }
      
      # measure the window contents
      @console_dlg.add_action_callback("windowInner") {|@console_dlg, param|
        # param = eval(param)
        # @windowBorderWidth = (@width-param[0].to_i)/2 if !@windowBorderWidth
        # if !@windowTitlebarHeight
          # @windowTitlebarHeight = (@height-param[1]).to_i
        # end
        # @console_dlg.execute_script("resizeHeight()")
      }
      
      # resize the window
      @console_dlg.add_action_callback("resizeHeight") {|@console_dlg, param|
        # param = eval(param)
        # @width = (2*@windowBorderWidth).to_i + param[0].to_i
        # @height = (@windowTitlebarHeight+param[1]).to_i - (2*@windowBorderWidth).to_i
        # @console_dlg.set_size(@width, @height)
        # Sketchup.write_default("Plugins_ae", "Console_Size",[@width,@height])
      }
      
      # when WebDialog is ready
      @console_dlg.add_action_callback("initialized") { |@console_dlg, param|
        @@open = true
      }
      
      # for debugging
      @console_dlg.add_action_callback("puts") { |@console_dlg, param|
        puts(param)
      }
      
      # get command from WebDialog for evaluation and send result back
      @console_dlg.add_action_callback("eval") { |@console_dlg, param|
        command = @console_dlg.get_element_value("returnvalue")
        begin
          result = eval(command, @binding)
          @console_dlg.execute_script( "result(" + result.inspect.inspect + ",\"result\"," + Time.now.to_i.to_s + ")")
        rescue Exception => e  
          @console_dlg.execute_script( "result(" + e.message.inspect + ",\"error\"," + Time.now.to_i.to_s + ")")
          @console_dlg.execute_script( "result(" + e.backtrace.inspect.inspect + ",\"error\"," + Time.now.to_i.to_s + ")")
        end
      }
      
      # get command from WebDialog for evaluation without return value
      @console_dlg.add_action_callback("do") { |@console_dlg, param|
        command = @console_dlg.get_element_value("returnvalue")
        begin
          eval(command, @binding)
        end
      }
      
      # cleanup, saving defaults, etc.
      @console_dlg.add_action_callback("do_on_close"){|@console_dlg, defaults|
        if @@debug
          puts("AE Console defaults:")
          puts(defaults.inspect)
        end
        #
        result = save_defaults(defaults)
        @@open = false
        UI.messagebox("AE::Console (debug)\n\nConsole has been closed.      \nSave defaults was '#{result}'\n") if @@debug
        #
        self.class.close() # calls the AE::Console class close() method.
      }
      
      # @console_dlg.set_on_close{
        # save_defaults
        # @@open = false
      # }
      
      (OSX)? @console_dlg.show_modal : @console_dlg.show
      
      return @console_dlg
      
    end # create()


    public


    def load_defaults
      return Sketchup.read_default("Plugins_ae", "Console", "{}")
    end


    def open
      dlg = @console_dlg
      if dlg.visible?
        dlg.bring_to_front
      else
        OSX ? dlg.show_modal() : dlg.show()
      end
    end


    def save_defaults(defaults)
      #@console_dlg.execute_script("saveDefaults()")
      #defaults = @console_dlg.get_element_value("returnvalue")
      # Replace all escape characters \ by a less used character (ie. &) to avoid multiple \\.
      # Escape all double quotes ".
      defaults = defaults.gsub(/\&/,'&amp;').gsub(/\\/,'&').gsub(/"/,'\"')
      Sketchup.write_default("Plugins_ae", "Console", defaults)
      # returns the true|false result of Sketchup.write_default()
    end


    def write(message)
      if @console_dlg && @@open
        if message.to_s[/^\n$/]
          # "puts" sends another separate \n afterwards
          @console_dlg.execute_script( "createNewRow()" ) 
        else
          @console_dlg.execute_script( "result(" + message.to_s.inspect + ", \"ruby\", " + Time.now.to_i.to_s + ")" )
        end
      end
      super(message) # call inherited class write function
    end
  #
  #}#


  #{#  RUN ONCE
  #
  unless file_loaded?(File.basename(__FILE__))
  
    @@command = UI::Command.new("Ruby Console (WebDialog)"){
      # in case you want to turn off traces
      $old_stdout = $stdout unless defined?($old_stdout)
      $old_stderr = $stderr unless defined?($old_stderr)
      #
      $stdout = AE::Console.new
      $stderr = $stdout # trap error displays as well
      self.open($stdout)
    }

    # Menu
    @@topmenu = UI.menu("Window")
    @@menuitem = @@topmenu.add_item(@@command)

    file_loaded(File.basename(__FILE__))

  end
  #
  #}#


end # class AE::Console



